-----------------------------------------------------------------------------

local AceOO = AceLibrary("AceOO-2.0")
local SML = AceLibrary("SharedMedia-1.0")
local Dewdrop = AceLibrary("Dewdrop-2.0")
local AssessmentBarInterface = Assessment:NewModule("BarInterface", "AceEvent-2.0")
AssessmentBarInterface.db = Assessment:AcquireDBNamespace("BarInterface")
Assessment:RegisterDefaults("BarInterface", "profile", {
	windows =
	{
-- Set on first login if it doesn't exist, and defaultWindowProfile copied in
--		count = 1,
	},
})

local AssessmentDisplaySets = Assessment:GetModule("DisplaySets")

-----------------------------------------------------------------------------

local SML = AceLibrary("SharedMedia-1.0")
local SML_statusbar = SML:List("statusbar")
local SML_font = SML:List("font")

local Dewdrop = AceLibrary:HasInstance("Dewdrop-2.0") and AceLibrary("Dewdrop-2.0")

-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentBarInterface")
L:RegisterTranslations("enUS", function() return {
	["Assessment"] = true,

	["Windows"] = true,
	["Window options."] = true,
	["Display Sets"] = true,
	["Display set options."] = true,
	["Add New Window"] = true,
	["Add a new assessment window to the user interface."] = true,

	["Window %d"] = true,
	["Options for assessment window %d"] = true,
	["Display Set"] = true,
	["Choose what to show for this window"] = true,
	["Visible"] = true,
	["Make the assessment window visible"] = true,

	["Display Set Data"] = true,
	["Display set data."] = true,

	["Lock"] = true,
	["Enable to prevent this window from being draggable."] = true,

	["Update Rate"] = true,
	["Controls the maximum number of times per second to update this window. Use 0 for continuous updates."] = true,

	["Backdrop Color"] = true,
	["Change the backdrop color and alpha for the window."] = true,
	["Highlight Color"] = true,
	["Change the highlight color and alpha for the window."] = true,
	["Border Style"] = true,
	["Change the border style for the window."] = true,
	["None"] = true,
	["Default"] = true,
	["Textured"] = true,
	["Square"] = true,
	["Border Color"] = true,
	["Change the border color and alpha for the window."] = true,

	["Bar Options"] = true,
	["Bar options for this window"] = true,
	["Bar Count"] = true,
	["Number of bars to show. Using this overrides the bar height setting."] = true,
	["Bar Height"] = true,
	["Height of bars."] = true,
	["Texture"] = true, 
	["Sets the bar texture"] = true,
	["Typeface"] = true,
	["Bar Typeface options"] = true,
	["Alpha"] = true,
	["Set the text alpha."] = true,
	["Font"] = true,
	["Font for text."] = true,
	["Font Size"] = true,
	["Font size for text."] = true,
	["Font Shadow"] = true,
	["Show font shadow on text"] = true,
	["Outline"] = true,
	["Sets the outline for text"] = true,
	["Segment Alpha"] = true,
	["Set the segment alpha."] = true,
	["Segment Texture"] = true, 
	["Sets the segment bar texture"] = true,

	["Title"] = true,
	["Title options."] = true,
	["Enable"] = true,
	["Enable title text."] = true,
	["Alignment"] = true,
	["Select how to align the title."] = true,
	["Left"] = true,
	["Center"] = true,
	["Right"] = true,
	["Font"] = true,
	["Font for title."] = true,
	["Font Size"] = true,
	["Font size for title."] = true,
	["Font Shadow"] = true,
	["Show font shadow on title text"] = true,

	["Outline"] = true,
	["Sets the outline for title text"] = true,
	["None"] = true,
	["Normal"] = true,
	["Thick"] = true,

	["Title Height"] = true,
	["Title height."] = true,
	["Background Color"] = true,
	["Change the title backdrop color."] = true,
	["Text Color"] = true,
	["Change the title text color."] = true,

	["This display set cannot be removed currently because one or more windows require its data."] = true,

	["Remove"] = true,
	["Remove this window from display"] = true,
	
	["ReportUsage1"] = "Usage: /asr {windowIndex}[channelCode]{limit} {channel/whisper name}",
	["ReportUsage2"] = "Channel codes are:",
	["ReportUsage3"] = "  C: Chat Frame",
	["ReportUsage4"] = "  G: Guild",
	["ReportUsage5"] = "  H: Channel",
	["ReportUsage6"] = "  O: Officer",
	["ReportUsage7"] = "  P: Party",
	["ReportUsage8"] = "  R: Raid",
	["ReportUsage9"] = "  S: Say",
	["ReportUsage10"] = "  W: Whisper",
	["ReportUsage11"] = " ",
	["ReportUsage12"] = "If windowIndex is omitted, the last window index is used.",
	["ReportUsage13"] = "If limit is omitted, all entries are shown",
	["ReportUsage14"] = "",
	["ReportUsage15"] = "Example: /asr 1w3 playerName",
	["ReportUsage16"] = "This will whisper playerName with the top 3 results from window 1.",
	
	["Parameters separated by \",\""] = true,


	["Display Set"] = true,
	["Set the root display set for this window."] = true,

	["Fight Data"] = true,
	["Select which fight data to show for this window."] = true,
	["Default"] = true,
	["Use the fight data specified by the display set."] = true,
	["Total Data"] = true,
	["Current Fight (1)"] = true,
	["Previous Fight (2)"] = true,
	["History #%d"] = true,

	["Save Data"] = true,
	["Save the data shown for future reference."] = true,

	["Report To"] = true,
	["Report Options"] = true,
	["Type a player's name."] = true,
	["Chat Frame"] = true,
	
	["Limit"] = true,
	["Select how many entries to show. Set to 0 to show all."] = true,
	
	-- Excluded channels for reporting list
	["General"] = true,
	["Trade"] = true,
	["LocalDefense"] = true,
	["WorldDefense"] = true,
	["GuildRecruitment"] = true,
} end)

-----------------------------------------------------------------------------

local defaultWindowProfile =
{
	fightIndex			= false,
	displaySetStack		= { "Assessment" },
	displaySetStackLevel= 1,
	updateRate			= 10,
	title = 
	{
		enable			= true,
		alignment		= "CENTER",
		font			= "Arial Narrow",
		fontSize		= 9,
		outline			= "OUTLINE",
		shadow			= true,
		height			= 10,
		textColor		= { 1, 1, 1, 1 },
		backgroundColor	= { 0, 0, 0, 1 }
	},
	bar = 
	{ 
		count			= 0,
		height			= 12,		-- Height is only used if count = 0
		texture			= "Blizzard",
		highlightColor	= { 1, 1, 0, 0.35 },
		font			= "Arial Narrow",
		fontSize		= 9,
		outline			= "",
		shadow			= true,
		textAlpha		= 1,
		segmentAlpha	= 1,
		segmentCount	= 10,
		segmentTexture	= "Blizzard",
	},
	frame = 
	{ 
		lock			= false,
		width			= 110,
		height			= 140, 
		point			= "CENTER",
		relativePoint	= "CENTER",
		x				= 0, 
		y				= 0, 
		borderStyle		= "Default",
		borderColor		= { 1, 1, 1, 0.75 },
		backdropColor	= { 0, 0, 0, 0.75 },
		visible			= true,
	},
}

-----------------------------------------------------------------------------

Assessment.Options.args.windows =
{
	type = "group",
	name = L["Windows"],
	desc = L["Window options."],
	order = 60,
	args = { },
}

local windowsArgs =
{ 
	add =
	{
		type = "execute",
		name = L["Add New Window"],
		desc = L["Add a new assessment window to the user interface."],
		func = function() AssessmentBarInterface:AddWindow() end,
		order = -1
	},
}

-----------------------------------------------------------------------------

local get = AssessmentBarInterface.GetOptionValue
local set = AssessmentBarInterface.SetOptionValue

local function windowGet(passValue)
	return get(AssessmentBarInterface.db.profile.windows[passValue[1]], passValue[2])
end

local function windowSet(passValue, value)
	set(AssessmentBarInterface.db.profile.windows[passValue[1]], passValue[2], value)
	AssessmentBarInterface:UpdateWindowOptions(passValue[1])
end

local function windowGetDisplaySet(passValue)
	local windowOptions = AssessmentBarInterface.db.profile.windows[passValue[1]]
	return ("$"):split(windowOptions.displaySetStack[windowOptions.displaySetStackLevel], 2)
end

local function windowSetDisplaySet(passValue, value)
	local windowOptions = AssessmentBarInterface.db.profile.windows[passValue[1]]
	
	windowOptions.displaySetStackLevel = 1
	windowOptions.displaySetStack[1] = value

	AssessmentBarInterface:UpdateWindowOptions(passValue[1])
end

local function windowGetDisplaySetData(passValue)
	local windowOptions = AssessmentBarInterface.db.profile.windows[passValue[1]]
	local _, result = ("$"):split(windowOptions.displaySetStack[windowOptions.displaySetStackLevel], 2)
	return result
end

local function windowSetDisplaySetData(passValue, value)
	local windowOptions = AssessmentBarInterface.db.profile.windows[passValue[1]]
	local displaySet = ("$"):split(windowOptions.displaySetStack[windowOptions.displaySetStackLevel], 2)
	
	windowOptions.displaySetStack[windowOptions.displaySetStackLevel] = displaySet.."$"..value

	AssessmentBarInterface:UpdateWindowOptions(passValue[1])
end

local function windowRemove(passValue)
	AssessmentBarInterface:RemoveWindow(passValue[1])
	Dewdrop:Close()
end

-----------------------------------------------------------------------------

local displaySetValidate		= { }

-----------------------------------------------------------------------------

local windowOptionsTemplate =
{
	type = "group",
--	name = ["Window %d"] = true,
--	desc = ["Options for assessment window %d"] = true,
	args =
	{
		visible =
		{
			type = "toggle",
			name = L["Visible"],
			desc = L["Make the assessment window visible"],
			get = windowGet,
			set = windowSet,
			passValue = "frame.visible",
			order = 1,
		},
		displaySet = 
		{
			type = "text",
			name = L["Display Set"],
			desc = L["Choose what to show for this window"],
			get = windowGetDisplaySet,
			set = windowSetDisplaySet,
			passValue = "displaySet",
			validate = displaySetValidate,
			order = 2,
		},
		displaySetData = 
		{
			type = "text",
			name = L["Display Set Data"],
			desc = L["Display set data."],
			get = windowGetDisplaySetData,
			set = windowSetDisplaySetData,
			passValue = "displaySet",
			order = 3,
			usage = L["Parameters separated by \",\""],
		},
		lock =
		{
			type = "toggle",
			name = L["Lock"],
			desc = L["Enable to prevent this window from being draggable."],
			get = windowGet,
			set = windowSet,
			passValue = "frame.lock",
		},
		backdropColor = 
		{
			type = "color",
			name = L["Backdrop Color"],
			desc = L["Change the backdrop color and alpha for the window."],
			get = function(passValue) return unpack(windowGet(passValue)) end,
			set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
			passValue = "frame.backdropColor",
			hasAlpha = true,
		},
		borderStyle = 
		{
			type = "text",
			name = L["Border Style"],
			desc = L["Change the border style for the window."],
			get = windowGet,
			set = windowSet,
			validate = { None = L["None"], Default = L["Default"], Textured = L["Textured"], Square = L["Square"] },
			passValue = "frame.borderStyle",
		},
		borderColor = 
		{
			type = "color",
			name = L["Border Color"],
			desc = L["Change the border color and alpha for the window."],
			get = function(passValue) return unpack(windowGet(passValue)) end,
			set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
			passValue = "frame.borderColor",
			hasAlpha = true,
		},
		updateRate =
		{
			type = "range",
			name = L["Update Rate"],
			desc = L["Controls the maximum number of times per second to update this window. Use 0 for continuous updates."],
			min  = 0,
			max  = 60,
			step = 1,
			passValue = "updateRate",
			get = windowGet,
			set = windowSet,
		},
		bars =
		{
			type = "group",
			name = L["Bar Options"],
			desc = L["Bar options for this window"],
			args =
			{
				count =
				{
					type = "range",
					name = L["Bar Count"],
					desc = L["Number of bars to show. Using this overrides the bar height setting."],
					min	 = 0,
					max  = 50,
					step = 1,
					get = windowGet,
					set = windowSet,
					passValue = "bar.count",
				},
				height =
				{
					type = "range",
					name = L["Bar Height"],
					desc = L["Height of bars."],
					min	 = 5,
					max  = 50,
					step = 1,
					disabled = function(passValue) return AssessmentBarInterface.db.profile.windows[passValue[1]].bar.count ~= 0 end,
					get = windowGet,
					set = windowSet,
					passValue = "bar.height",
				},
				highlightColor = 
				{
					type = "color",
					name = L["Highlight Color"],
					desc = L["Change the highlight color and alpha for the window."],
					get = function(passValue) return unpack(windowGet(passValue)) end,
					set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
					passValue = "bar.highlightColor",
					hasAlpha = true,
				},
				texture =
				{
					type = "text", 
					name = L["Texture"], 
					desc = L["Sets the bar texture"],
					get = windowGet,
					set = windowSet,
					validate = SML_statusbar,
					passValue = "bar.texture",
				},
				segmentAlpha =
				{
					type = "range",
					name = L["Segment Alpha"],
					desc = L["Set the segment alpha."],
					min = 0,
					max = 1,
					step = 0.01,
					get = windowGet,
					set = windowSet,
					passValue = "bar.segmentAlpha",
				},
				segmentTexture =
				{
					type = "text", 
					name = L["Segment Texture"], 
					desc = L["Sets the segment bar texture"],
					get = windowGet,
					set = windowSet,
					validate = SML_statusbar,
					passValue = "bar.segmentTexture",
				},
				typeface =
				{
					type = "group",
					name = L["Typeface"],
					desc = L["Bar Typeface options"],
					args =
					{
						alpha =
						{
							type = "range",
							name = L["Alpha"],
							desc = L["Set the text alpha."],
							min = 0,
							max = 1,
							step = 0.01,
							get = windowGet,
							set = windowSet,
							passValue = "bar.textAlpha",
						},
						font =
						{
							type = "text",
							name = L["Font"],
							desc = L["Font for text."],
							get = windowGet,
							set = windowSet,
							passValue = "bar.font",
							validate = SML_font,
						},
						fontSize =
						{
							type = "range",
							name = L["Font Size"],
							desc = L["Font size for text."],
							min = 4,
							max = 32,
							step = 1,
							get = windowGet,
							set = windowSet,
							passValue = "bar.fontSize",
						},
						shadow =
						{
							type = "toggle",
							name = L["Font Shadow"],
							desc = L["Show font shadow on text"],
							get = windowGet,
							set = windowSet,
							passValue = "bar.shadow",
						},
						outline =
						{
							type = "text",
							name = L["Outline"],
							desc = L["Sets the outline for text"],
							get = windowGet,
							set = windowSet,
							passValue = "bar.outline",
							validate = { [""] = L["None"], OUTLINE = L["Normal"], THICKOUTLINE = L["Thick"] },
						},
					}
				},
			},
		},
		title =
		{
			type = "group",
			name = L["Title"],
			desc = L["Title options."],
			args =
			{
				enable = 
				{
					type = "toggle",
					name = L["Enable"],
					desc = L["Enable title text."],
					get = windowGet,
					set = windowSet,
					passValue = "title.enable",
					order = 1,
				},
				alignment =
				{
					type = "text",
					name = L["Alignment"],
					desc = L["Select how to align the title."],
					get = windowGet,
					set = windowSet,
					passValue = "title.alignment",
					validate = { LEFT = L["Left"], CENTER = L["Center"], RIGHT = L["Right"] },
				},
				font =
				{
					type = "text",
					name = L["Font"],
					desc = L["Font for title."],
					get = windowGet,
					set = windowSet,
					passValue = "title.font",
					validate = SML_font,
				},
				height =
				{
					type = "range",
					name = L["Title Height"],
					desc = L["Title height."],
					min  = 0,
					max  = 32,
					step = 1,
					get = windowGet,
					set = windowSet,
					passValue = "title.height",
				},
				fontSize =
				{
					type = "range",
					name = L["Font Size"],
					desc = L["Font size for title."],
					min = 4,
					max = 32,
					step = 1,
					get = windowGet,
					set = windowSet,
					passValue = "title.fontSize",
				},
				shadow =
				{
					type = "toggle",
					name = L["Font Shadow"],
					desc = L["Show font shadow on title text"],
					get = windowGet,
					set = windowSet,
					passValue = "title.shadow",
				},
				backgroundColor = 
				{
					type = "color",
					name = L["Background Color"],
					desc = L["Change the title backdrop color."],
					get = function(passValue) return unpack(windowGet(passValue)) end,
					set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
					passValue = "title.backgroundColor",
					hasAlpha = true,
				},
				textColor = 
				{
					type = "color",
					name = L["Text Color"],
					desc = L["Change the title text color."],
					get = function(passValue) return unpack(windowGet(passValue)) end,
					set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
					passValue = "title.textColor",
					hasAlpha = true,
				},
				outline =
				{
					type = "text",
					name = L["Outline"],
					desc = L["Sets the outline for title text"],
					get = windowGet,
					set = windowSet,
					passValue = "title.outline",
					validate = { [""] = L["None"], OUTLINE = L["Normal"], THICKOUTLINE = L["Thick"] },
				},
			},
		},
		remove =
		{
			type = "execute",
			name = L["Remove"],
			desc = L["Remove this window from display"],
			func = windowRemove,
			passValue = false,
			order = -1,
		},
	}
}

-----------------------------------------------------------------------------

local lastReportWhisperTarget = ""
local reportLimit = 0
local function windowReport(data)
	AssessmentBarInterface.windows[data[1]]:Report(data[2][1], reportLimit, data[2][2])
end

local function windowReportWhisper(data, target)
	if target == "" then return end
	lastReportWhisperTarget = target
	AssessmentBarInterface.windows[data[1]]:Report("WHISPER", reportLimit, lastReportWhisperTarget)
	Dewdrop:Close()
end

local function windowFightDataGet(passValue)
	return AssessmentBarInterface.db.profile.windows[passValue[1]].fightIndex == passValue[2]
end

local function windowFightDataSet(passValue, value)
	AssessmentBarInterface.db.profile.windows[passValue[1]].fightIndex = passValue[2]
	AssessmentBarInterface:UpdateWindowOptions(passValue[1])
end

local function saveFightData(passValue)
	Dewdrop:Close()
	StaticPopupDialogs["ASSESSMENT_SAVE_FIGHT_DATA"].data = AssessmentBarInterface.windows[passValue[1]].fightIndex
	StaticPopup_Show("ASSESSMENT_SAVE_FIGHT_DATA")
end

local windowMenuTemplate =
{
	type = "group",
	name = "",
	desc = "",
	args =
	{
		report =
		{
			type = "group",
			name = L["Report To"],
			desc = L["Report Options"],
			args =
			{
				raid =
				{
					type = "execute",
					name = CHAT_MSG_RAID,
					desc = CHAT_MSG_RAID,
					func = windowReport,
					hidden = function() return GetNumRaidMembers() == 0 end,
					passValue = { "RAID" },
					order = 1,
				},
				guild =
				{
					type = "execute",
					name = CHAT_MSG_GUILD,
					desc = CHAT_MSG_GUILD,
					func = windowReport,
					hidden = function() return not IsInGuild() end,
					passValue = { "GUILD" },
					order = 2,
				},
				party =
				{
					type = "execute",
					name = CHAT_MSG_PARTY,
					desc = CHAT_MSG_PARTY,
					func = windowReport,
					hidden = function() return GetNumPartyMembers() == 0 end,
					passValue = { "PARTY" },
					order = 3,
				},
				say =
				{
					type = "execute",
					name = CHAT_MSG_SAY,
					desc = CHAT_MSG_SAY,
					func = windowReport,
					passValue = { "SAY" },
				},
				officer =
				{
					type = "execute",
					name = CHAT_MSG_OFFICER,
					desc = CHAT_MSG_OFFICER,
					func = windowReport,
					hidden = function() return not IsInGuild() end,
					passValue = { "OFFICER" },
				},
				chatFrame =
				{
					type = "group",
					name = L["Chat Frame"],
					desc = L["Chat Frame"],
					args =
					{
						chatFrame1 =
						{
							type = "execute",
							name = L["Chat Frame"].." 1",
							desc = L["Chat Frame"].." 1",
							func = windowReport,
							passValue = { "CHATFRAME", "ChatFrame1" },
						},
						chatFrame2 =
						{
							type = "execute",
							name = L["Chat Frame"].." 2",
							desc = L["Chat Frame"].." 2",
							func = windowReport,
							passValue = { "CHATFRAME", "ChatFrame2" },
						},
						chatFrame3 =
						{
							type = "execute",
							name = L["Chat Frame"].." 3",
							desc = L["Chat Frame"].." 3",
							func = windowReport,
							passValue = { "CHATFRAME", "ChatFrame3" },
						},
						chatFrame4 =
						{
							type = "execute",
							name = L["Chat Frame"].." 4",
							desc = L["Chat Frame"].." 4",
							func = windowReport,
							passValue = { "CHATFRAME", "ChatFrame4" },
						},
						chatFrame5 =
						{
							type = "execute",
							name = L["Chat Frame"].." 5",
							desc = L["Chat Frame"].." 5",
							func = windowReport,
							passValue = { "CHATFRAME", "ChatFrame5" },
						},
						chatFrame6 =
						{
							type = "execute",
							name = L["Chat Frame"].." 6",
							desc = L["Chat Frame"].." 6",
							func = windowReport,
							passValue = { "CHATFRAME", "ChatFrame6" },
						},
						chatFrame7 =
						{
							type = "execute",
							name = L["Chat Frame"].." 7",
							desc = L["Chat Frame"].." 7",
							func = windowReport,
							passValue = { "CHATFRAME", "ChatFrame7" },
						},
					}
				},
				channels =
				{
					type = "group",
					name = CHANNELS,
					desc = CHANNELS,
					args =
					{
					},
				},
				whisper =
				{
					type = "text",
					name = CHAT_MSG_WHISPER_INFORM,
					desc = CHAT_MSG_WHISPER_INFORM,
					usage = L["Type a player's name."],
					get = function() return lastReportWhisperTarget end,
					set = windowReportWhisper,
					passValue = "WHISPER",
				},
				limit =
				{
					type = "range",
					name = L["Limit"],
					desc = L["Select how many entries to show. Set to 0 to show all."],
					min = 0,
					max = 40,
					step = 1,
					get = function() return reportLimit end,
					set = function(v) reportLimit = v end,
					order = -1,
				},
			},
		},
		displaySet =
		{
			type = "text",
			name = L["Display Set"],
			desc = L["Set the root display set for this window."],
			get = windowGetDisplaySet,
			set = windowSetDisplaySet,
			passValue = false,
			validate = { }
		},
		fightData =
		{
			type = "group",
			name = L["Fight Data"],
			desc = L["Select which fight data to show for this window."],
			args =
			{
				default =
				{
					type = "toggle",
					name = L["Default"],
					desc = L["Use the fight data specified by the display set."],
					get = windowFightDataGet,
					set = windowFightDataSet,
					passValue = false,
					order = 1,
				},
				spacer1 =
				{
					type = "header",
					order = 2,
				},
				history0 =
				{
					type = "toggle",
					name = L["Total Data"],
					desc = L["Total Data"],
					get = windowFightDataGet,
					set = windowFightDataSet,
					passValue = 0,
					order = 3,
				},
				history1 =
				{
					type = "toggle",
					name = L["Current Fight (1)"],
					desc = L["Current Fight (1)"],
					get = windowFightDataGet,
					set = windowFightDataSet,
					passValue = 1,
					order = 4,
				},
				history2 =
				{
					type = "toggle",
					name = L["Previous Fight (2)"],
					desc = L["Previous Fight (2)"],
					get = windowFightDataGet,
					set = windowFightDataSet,
					passValue = 2,
					order = 5,
				},
			},
		},
		saveFightData =
		{
			type = "execute",
			name = L["Save Data"],
			desc = L["Save the data shown for future reference."],
			func = saveFightData,
			passValue = false,
		},
		remove =
		{
			type = "execute",
			name = L["Remove"],
			desc = L["Remove this window from display"],
			func = windowRemove,
			passValue = false,
			order = -2,
		},
		lock =
		{
			type = "toggle",
			name = L["Lock"],
			desc = L["Enable to prevent this window from being draggable."],
			get = windowGet,
			set = windowSet,
			passValue = "frame.lock",
			order = -1,
		},
	}
}

-----------------------------------------------------------------------------

local listCompileData = 
{ 
	window =
	{
	}
}

local listCompileDataWindow = listCompileData.window

local function CreateParametersList(...)
	local t = AssessmentBarInterface:NewTable()	
	for i=1,select('#', ...) do
		t[i] = select(i, ...)
	end
	return t
end

-----------------------------------------------------------------------------

-- This thing is created on demand, and reparented as necessary
local uiBarHighlightFrame

local UIBar = AceOO.Class()

local function OnUIBarEnter(self, motion)
	local frame = self.uiBar.frame
	if not uiBarHighlightFrame then
		uiBarHighlightFrame = frame:CreateTexture(nil, "HIGHLIGHT")
		uiBarHighlightFrame:SetTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight")
		uiBarHighlightFrame:SetBlendMode("ADD")
	end
	uiBarHighlightFrame:SetVertexColor(unpack(self.uiBar.windowOptions.bar.highlightColor))
	uiBarHighlightFrame:SetParent(frame)
	uiBarHighlightFrame:ClearAllPoints()
	uiBarHighlightFrame:SetAllPoints(frame)
	uiBarHighlightFrame:Show()
end

local function OnUIBarLeave(self, motion)
	uiBarHighlightFrame:Hide()
end

local function OnUIBarMouseDown(self, button)
	if button == "LeftButton" then
		self:GetParent().window:OnBarClicked(self.uiBar)
	elseif button == "RightButton" then
		if IsShiftKeyDown() or IsControlKeyDown() or IsAltKeyDown() then
			self:GetParent().window:OnClick(button)
		else
			self:GetParent().window:PopDisplaySetStack()
		end
	end
end

function UIBar.prototype:init(parent, window, index)
	UIBar.super.prototype.init(self)

	local frame = CreateFrame("Frame", nil, parent)
	frame.uiBar = self
	self.frame = frame

	local bar = frame:CreateTexture()
	self.bar = bar
	self.texts = { }

	bar:SetPoint("TOPLEFT", frame, "TOPLEFT")
	bar:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT")
	
	frame:SetScript("OnEnter", OnUIBarEnter)
	frame:SetScript("OnLeave", OnUIBarLeave)
	frame:SetScript("OnMouseDown", OnUIBarMouseDown)
	frame:EnableMouse()

	self:SetOptions(parent, window, index)
end

function UIBar.prototype:SetOptions(parent, window, index)
	self.windowOptions = window.windowOptions
	self.displaySet = window.displaySet

	local bar = self.bar

	bar:SetTexture(SML:Fetch("statusbar", self.windowOptions.bar.texture))
	bar:SetVertexColor(1, 1, 1, 1)

	for i = 1, self.displaySet.text.count do
		if self.displaySet.text[i].visible then
			if not self.texts[i] then
				self.texts[i] = self.frame:CreateFontString(nil, "OVERLAY")
			end
			self.texts[i]:Show()
		elseif self.texts[i] then
			self.texts[i]:Hide()
		end
	end

	for i = self.displaySet.text.count + 1, #self.texts do
		self.texts[i]:Hide()
	end
	
	if window.segmentValueMethod then
		self.barSegments = self.barSegments or { }
		for i = 1, self.windowOptions.bar.segmentCount do
			local barSegment = self.barSegments[i]
			if not barSegment then
				barSegment = self.frame:CreateTexture(nil, "OVERLAY")
				barSegment:SetVertexColor(1, 1, 1, 1)
				barSegment:ClearAllPoints()
				barSegment:SetPoint("TOP", self.frame, "TOP")
				barSegment:SetPoint("BOTTOM", self.frame, "BOTTOM")
				
				if i == 1 then
					barSegment:SetPoint("LEFT", self.frame, "LEFT")
				else
					barSegment:SetPoint("LEFT", self.barSegments[i-1], "RIGHT")
				end
				
				self.barSegments[i] = barSegment
			end
			barSegment:SetTexture(SML:Fetch("statusbar", self.windowOptions.bar.segmentTexture))
			barSegment:SetAlpha(self.windowOptions.bar.segmentAlpha)
			barSegment:Hide()
		end
		for i = self.windowOptions.bar.segmentCount+1, #self.barSegments do
			self.barSegments[i]:Hide()
		end
	elseif self.barSegments then
		for i,b in ipairs(self.barSegments) do
			b:Hide()
		end
	end
	self.numberOfVisibleBarSegments = 0
	
	self:PlaceBar(parent, index)
end

function UIBar.prototype:PlaceBar(parent, index)
	local frame = self.frame
	local bar = self.bar
	local verticalOffset = -5 - (self.windowOptions.title.enable and self.windowOptions.title.height or 0)
	local height = parent.window.barHeight
	local barOptions = self.windowOptions.bar
	
	self.maximumBarWidth = (self.windowOptions.frame.width-10)
	
	frame:ClearAllPoints()
	frame:SetPoint("TOPLEFT", parent, "TOPLEFT", 5, -(index-1)*height+verticalOffset)
	frame:SetPoint("BOTTOMRIGHT", parent, "TOPRIGHT", -5, -index*height+verticalOffset)
	
	for i = 1, self.displaySet.text.count do
		local text = self.texts[i]
		if text then
			local options = self.displaySet.text[i]
			
			text:ClearAllPoints()
			text:SetPoint("TOPLEFT", frame, "TOPLEFT", options.offsets.left, options.offsets.vertical)
			text:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", options.offsets.right, options.offsets.vertical)
			text:SetJustifyH(options.alignment)

			text:SetAlpha(barOptions.textAlpha)
			text:SetFont(SML:Fetch("font", barOptions.font), barOptions.fontSize, barOptions.outline)
			if barOptions.shadow then
				text:SetShadowOffset(1, -1)
			else
				text:SetShadowOffset(0, 0)
			end
			text:SetText("")
		end
	end
end

function UIBar.prototype:Hide()
	self.frame:Hide()
end

function UIBar.prototype:HideAllSegments()
	for i = 1, self.numberOfVisibleBarSegments do
		self.barSegments[i]:Hide()
	end
	self.numberOfVisibleBarSegments = 0
end

function UIBar.prototype:Update(data, barNormalizeMultiplier, window)
	-- Update Bar
	local bar = self.bar
	local barWidth = data.barValue * barNormalizeMultiplier * self.maximumBarWidth
	if barWidth == 0 then
		bar:Hide()
		if window.segmentValueMethod then
			self:HideAllSegments()
		end
	else
		bar:SetWidth(barWidth)
		bar:Show()

		local color
		for i,f in ipairs(window.barColorMethods) do
			color = f(data.fightData, data)
			if color then break end
		end
		bar:SetVertexColor(unpack(color))

		if window.segmentValueMethod then
			local fightData = Assessment:GetFightData(window.segmentFightIndex)
			
			if fightData then
				local totalBarValue = 0
				local sortList = AssessmentBarInterface:NewTable()
				
				local windowParameters = CreateParametersList((","):split(window.segmentDataMethod(data)))
				listCompileDataWindow.windowParameters = windowParameters
				
				local listDataParameters = CreateParametersList((","):split(window.segmentListDataMethod(listCompileData)))
				listCompileDataWindow.listDataParameters = listDataParameters
				
				for k,v in window.segmentListTypeMethod(fightData, unpack(listDataParameters) ) do
					local segmentData = AssessmentBarInterface:NewTable()
					segmentData.key = k
					segmentData.value = v
					segmentData.fightData = fightData
					segmentData.window = listCompileDataWindow
					
					local barValue = window.segmentValueMethod(segmentData)
					segmentData.barValue = barValue

					if barValue > 0 and window.segmentFilterMethod(segmentData) then
						totalBarValue = totalBarValue + barValue

						local rank = #sortList+1
						segmentData.rank = rank
						sortList[rank] = segmentData
					else
						AssessmentBarInterface:ReleaseTable(segmentData)
					end			
				end
				
				listCompileDataWindow.listDataParameters = AssessmentBarInterface:ReleaseTable(listDataParameters)
				listCompileDataWindow.windowParameters = AssessmentBarInterface:ReleaseTable(windowParameters)
				
				if totalBarValue > 0 then
					for _,segmentData in ipairs(sortList) do
						segmentData.sortValue = window.segmentSortMethod(segmentData)
					end
				
					table.sort(sortList, window.segmentSortDirectionMethod)
					
					local newNumberOfSegments = #sortList
					if newNumberOfSegments > self.windowOptions.bar.segmentCount then
						newNumberOfSegments = self.windowOptions.bar.segmentCount
					end
					
					for i=1, newNumberOfSegments do
						local barSegment = self.barSegments[i]
						local segmentData = sortList[i]
						
						barSegment:SetWidth(segmentData.barValue * barWidth / totalBarValue)
						color = nil
						for _,f in ipairs(window.segmentBarColorMethods) do
							color = f(fightData, segmentData)
							if color then break end
						end
						barSegment:SetVertexColor(unpack(color))
						barSegment:Show()
					end
					
					for i=newNumberOfSegments+1, self.numberOfVisibleBarSegments do
						self.barSegments[i]:Hide()
					end
					self.numberOfVisibleBarSegments = newNumberOfSegments
				else
					self:HideAllSegments()
				end

				for i,segmentData in ipairs(sortList) do
					AssessmentBarInterface:ReleaseTable(segmentData)
				end
				AssessmentBarInterface:ReleaseTable(sortList)
			else
				self:HideAllSegments()
			end
		end
	end
	
	-- Update Texts
	local textMethods = window.textMethods
	local displaySetText = self.displaySet.text
	for i = 1, displaySetText.count do
		if displaySetText[i].visible then
			local text = self.texts[i]
			text:SetText(textMethods[i](data))
		end
	end
	self.frame:Show()
end

-----------------------------------------------------------------------------

local borderStyles =
{
	None		= { edgeFile = "", inset = 5 },
	Default		= { edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", inset = 5 },
	Textured	= { edgeFile = "", inset = 3 },
	Square		= { edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border", inset = 5 },
}

local UIWindow = AceOO.Class("AceEvent-2.0")

local bgTable = 
{
    bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
    tile = true,
    tileSize = 16,
    edgeSize = 16,
	insets = { left = inset, right = inset, top = inset, bottom = inset }
}

local isWindowSizing, wasWindowDragged, wasWindowSizeChanged

local function OnUIWindowMouseDown(this, button)
	if button == "LeftButton" then
		if not this.window.windowOptions.frame.lock then 
			local mouseDownX, mouseDownY = GetCursorPosition(UIParent);

			local effectiveScale = this:GetEffectiveScale();
			local left = this:GetLeft() * effectiveScale;
			local top = this:GetTop() * effectiveScale;
			local right = this:GetRight() * effectiveScale;
			local bottom = this:GetBottom() * effectiveScale;
			
			local edge = "";
			if (abs(mouseDownY - bottom) < 7) then
				edge = "BOTTOM";
			elseif (abs(mouseDownY - top) < 7) then
				edge = "TOP"
			end
			
			if (abs(mouseDownX - left) < 7) then
				edge = edge.."LEFT";
			elseif (abs(mouseDownX - right) < 7) then
				edge = edge.."RIGHT";
			end
			
			if edge ~= "" then
				this:StartSizing(edge)
				isWindowSizing = true
				wasWindowSizeChanged = nil
			end
		end
	elseif button == "RightButton" then
		if IsShiftKeyDown() or IsControlKeyDown() or IsAltKeyDown() then
			this.window:OnClick(button)
		else
			this.window:PopDisplaySetStack()
		end
	end
end

local function OnUIWindowMouseUp(this, button)
	if button ~= "LeftButton" then return end
	
	if isWindowSizing then
		this:StopMovingOrSizing() 
		isWindowSizing = nil
		if wasWindowSizeChanged then
			wasWindowSizeChanged = nil
		else
			this.window:OnClick(button)
		end
	elseif wasWindowDragged then
		wasWindowDragged = nil
	else
		this.window:OnClick(button)
	end
end

local function OnUIWindowDragStart(this)
	wasWindowDragged = true
	if not this.window.windowOptions.frame.lock then 
		this:StartMoving() 
	end
end

local function OnUIWindowDragStop(this)
	this:StopMovingOrSizing() 
	local point,_,relativePoint,x,y = this:GetPoint(1)
	local frameOptions = this.window.windowOptions.frame
	frameOptions.point = point
	frameOptions.relativePoint = relativePoint
	frameOptions.x = x
	frameOptions.y = y
end

local function OnUIWindowSizeChanged(this, width, height)
	wasWindowSizeChanged = true
	local window = this.window
	local frameOptions = window.windowOptions.frame
	frameOptions.width = width 
	frameOptions.height = height

	local point,_,relativePoint,x,y = this:GetPoint(1)
	frameOptions.point = point
	frameOptions.relativePoint = relativePoint
	frameOptions.x = x
	frameOptions.y = y

	window:LayoutBars()
	window.forceUpdate = true
	window:Update()
end

local function OnUIWindowMouseWheel(this, delta) 
	local window = this.window
	if delta < 0 then
		if IsControlKeyDown() then
			window.topBarOffset = window.topBarOffset + window.barCount
		else
			window.topBarOffset = window.topBarOffset + 1
		end
	else
		if IsControlKeyDown() then
			window.topBarOffset = window.topBarOffset - window.barCount
		else
			window.topBarOffset = window.topBarOffset - 1
		end
		if window.topBarOffset < 0 then
			window.topBarOffset = 0
		end
	end
	window.forceUpdate = true
	window:Update()
end

local function DescendingSortMethod(a, b) 
	if a.sortValue == b.sortValue then
		return a.key < b.key	-- always sort keys ascending if the sort values are equal
	else
		return a.sortValue > b.sortValue
	end
end

local function AscendingSortMethod(a, b) 
	if a.sortValue == b.sortValue then
		return a.key < b.key
	else
		return a.sortValue < b.sortValue
	end
end

function UIWindow.prototype:init(windowOptions, displaySet)
	UIWindow.super.prototype.init(self)
	local frame = CreateFrame("Frame", nil, UIParent)
	frame.window = self

	self.dependentData		= { }
	self.dependentEvents	= { }
	self.overrideEvents		= { }
	self.dataUpdateTimes	= { }
	self.textMethods		= { }
	self.windowParameters	= { }
	self.listDataParameters = { }
	self.barColorMethods	= { }
	self.segmentBarColorMethods	= { }

	self.frame			= frame
	self.bars			= { }
	
	self.titleText = frame:CreateFontString()
	self.titleTexture = frame:CreateTexture()

	frame:SetMovable(true)
	frame:SetResizable(true)
	frame:SetMinResize(32,32)
	frame:EnableMouse(true)
	frame:EnableMouseWheel(true)
	frame:RegisterForDrag("LeftButton")

	frame:SetScript("OnMouseWheel", OnUIWindowMouseWheel)
	frame:SetScript("OnMouseDown",	OnUIWindowMouseDown)
	frame:SetScript("OnMouseUp",	OnUIWindowMouseUp)
	frame:SetScript("OnDragStart",	OnUIWindowDragStart)
	frame:SetScript("OnDragStop",	OnUIWindowDragStop)
	frame:SetScript("OnSizeChanged",OnUIWindowSizeChanged)

	self:SetWindowOptions(windowOptions)
end

function returnDefaultBarColor(fightData, data)
	return data.window.defaultBarColor
end

function UIWindow.prototype:SetWindowOptions(windowOptions)
	local displaySet, displaySetParameterText = AssessmentBarInterface:GetDisplaySet(windowOptions)

	-- Clear some tables
	for k in pairs(self.dependentData) do self.dependentData[k] = nil end
	for k in pairs(self.dependentEvents) do self.dependentEvents[k] = nil end
	for k in pairs(self.overrideEvents) do self.overrideEvents[k] = nil end
	for k in pairs(self.dataUpdateTimes) do self.dataUpdateTimes[k] = nil end
	for k in ipairs(self.textMethods) do self.textMethods[k] = nil end

	self.topBarOffset	= 0
	self.forceUpdate	= true
	self.displaySet		= displaySet
	self.windowOptions	= windowOptions

	-- Data for this window
	AssessmentBarInterface:ReleaseTable(self.windowParameters)
	self.windowParameters = CreateParametersList((","):split(displaySetParameterText or ""))

	listCompileDataWindow.windowParameters = self.windowParameters
	local listDataText = Assessment:CompileTag(displaySet.listDataTag, "string").method(listCompileData)
	listCompileDataWindow.windowParameters = nil
	
	AssessmentBarInterface:ReleaseTable(self.listDataParameters)
	self.listDataParameters = CreateParametersList((","):split(listDataText))
	
	-- ListType
	local listType = Assessment.ListType[displaySet.listType]
	self.listTypeMethod = listType.method
	self:MergeRequiredData(listType.data, listType.events, listType.overrideEvents)

	-- Filter Tag
	local compiledFilterValue = Assessment:CompileTag(displaySet.filterTag, "raw")
	self:MergeRequiredData(compiledFilterValue.data, compiledFilterValue.events, compiledFilterValue.overrideEvents)
	self.filterMethod = compiledFilterValue.method

	-- Determine Fight Index
	self.fightIndex = windowOptions.fightIndex
	if not self.fightIndex then
		self.fightIndex = 0
		for i = windowOptions.displaySetStackLevel, 1, -1 do
			local displaySetName = ("$"):split(windowOptions.displaySetStack[i], 2)
			local set = AssessmentDisplaySets:GetDisplaySet(displaySetName)
			if not set then break end
			if set.fightIndex ~= -1 then
				self.fightIndex = set.fightIndex
				break
			end
		end
	end

	-- Bar Display Stuff

	self.sortDirectionMethod = displaySet.descendingSort and DescendingSortMethod or AscendingSortMethod

	local compiledBarValue = Assessment:CompileTag(displaySet.barValueTag, "number")
	self.barValueMethod = compiledBarValue.method
	self:MergeRequiredData(compiledBarValue.data, compiledBarValue.events, compiledBarValue.overrideEvents)
	
	for i=1,displaySet.text.count do
		local compiledTextValue = Assessment:CompileTag(displaySet.text[i].tag, "string")
		self.textMethods[i] = compiledTextValue.method
		if displaySet.text[i].visible then
			self:MergeRequiredData(compiledTextValue.data, compiledTextValue.events, compiledTextValue.overrideEvents)
		end
	end

	local compiledSortValue = Assessment:CompileTag(displaySet.sortKeyTag, "raw")
	self:MergeRequiredData(compiledSortValue.data, compiledSortValue.events, compiledSortValue.overrideEvents)
	self.sortMethod = compiledSortValue.method

	-------

	for i in ipairs(self.barColorMethods) do
		self.barColorMethods[i] = nil
	end
	for i = 1, displaySet.colorTypeList.count do
		local colorType = Assessment.ColorType[displaySet.colorTypeList[i]]
		self.barColorMethods[i] = colorType.method
		self:MergeRequiredData(colorType.data, colorType.events, colorType.overrideEvents)
	end
	self.barColorMethods[#self.barColorMethods+1] = returnDefaultBarColor
	self.defaultBarColor = displaySet.defaultColor

	-- Segment stuff
	self.segmentDataMethod = nil
	self.segmentValueMethod = nil
	self.segmentListTypeMethod = nil
	if displaySet.segmentEnable and displaySet.segmentDisplaySet then
		local segmentDisplaySet = AssessmentDisplaySets:GetDisplaySet(displaySet.segmentDisplaySet)
		if segmentDisplaySet then
			for i in ipairs(self.segmentBarColorMethods) do
				self.segmentBarColorMethods[i] = nil
			end
			for i = 1, segmentDisplaySet.colorTypeList.count do
				self.segmentBarColorMethods[i] = Assessment.ColorType[segmentDisplaySet.colorTypeList[i]].method
			end
			self.segmentBarColorMethods[#self.segmentBarColorMethods+1] = returnDefaultBarColor
			listCompileDataWindow.defaultBarColor = segmentDisplaySet.defaultColor

			self.segmentDataMethod = Assessment:CompileTag(displaySet.segmentData, "string").method
			self.segmentValueMethod = Assessment:CompileTag(segmentDisplaySet.barValueTag, "number").method
			self.segmentListTypeMethod = Assessment.ListType[segmentDisplaySet.listType].method
			self.segmentListDataMethod = Assessment:CompileTag(segmentDisplaySet.listDataTag, "string").method
			self.segmentFilterMethod = Assessment:CompileTag(segmentDisplaySet.filterTag, "raw").method
			self.segmentFightIndex = windowOptions.fightIndex
			if not self.segmentFightIndex then
				self.segmentFightIndex = segmentDisplaySet.fightIndex
				if self.segmentFightIndex == -1 then
					self.segmentFightIndex = self.fightIndex
				end
			end
			self.segmentSortMethod = Assessment:CompileTag(segmentDisplaySet.sortKeyTag, "raw").method
			self.segmentSortDirectionMethod = segmentDisplaySet.descendingSort and DescendingSortMethod or AscendingSortMethod
		end
	end

	-------

	local frame = self.frame
	local frameOptions = windowOptions.frame
	frame:ClearAllPoints()
	frame:SetPoint(frameOptions.point, UIParent, frameOptions.relativePoint, frameOptions.x, frameOptions.y)
	frame:SetWidth(frameOptions.width)
	frame:SetHeight(frameOptions.height)

	-------

	local borderStyle = borderStyles[windowOptions.frame.borderStyle]
	bgTable.edgeFile = borderStyle.edgeFile
	local inset = borderStyle.inset
	local insets = bgTable.insets
	insets.left = inset
	insets.top = inset
	insets.right = inset
	insets.bottom = inset
    frame:SetBackdrop(bgTable)
	frame:SetBackdropBorderColor(unpack(windowOptions.frame.borderColor))
	frame:SetBackdropColor(unpack(windowOptions.frame.backdropColor))

	--------

	local titleText = self.titleText
	local titleOptions = windowOptions.title
	
	if titleOptions.enable then
		titleText:ClearAllPoints()
		titleText:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5)
		titleText:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -5, -titleOptions.height-5)
		titleText:SetTextColor(unpack(titleOptions.textColor))

		titleText:SetFont(SML:Fetch("font", titleOptions.font), titleOptions.fontSize, titleOptions.outline)
		if titleOptions.shadow then
			titleText:SetShadowOffset(1, -1)
		else
			titleText:SetShadowOffset(0, 0)
		end
		titleText:SetJustifyH(titleOptions.alignment)

		local compiledTitleTextValue = Assessment:CompileTag(displaySet.titleTag, "string")
		self.titleTextMethod = compiledTitleTextValue.method
		self:MergeRequiredData(compiledTitleTextValue.data, compiledTitleTextValue.events, compiledTitleTextValue.overrideEvents)

		titleText:Show()

		local titleTexture = self.titleTexture
		titleTexture:ClearAllPoints()
		titleTexture:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5)
		titleTexture:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -5, -titleOptions.height-5)
		titleTexture:SetTexture(unpack(titleOptions.backgroundColor))
		titleTexture:Show()
	else
		titleText:Hide()
		self.titleTexture:Hide()
	end

	--------
	
	for k in pairs(self.overrideEvents) do
		self.dependentEvents[k] = nil
	end
	
	--------

	self:LayoutBars()

	if frameOptions.visible then
		self:Show()
	else
		self:Hide()
	end
end

function UIWindow.prototype:LayoutBars()
	local barOptions = self.windowOptions.bar
	
	if barOptions.count == 0 then
		self.barHeight = barOptions.height
		self.barCount = floor((self.frame:GetHeight()-10-(self.windowOptions.title.enable and self.windowOptions.title.height or 0)) / barOptions.height)
	else
		self.barCount = barOptions.count
		self.barHeight = (self.frame:GetHeight()-10-(self.windowOptions.title.enable and self.windowOptions.title.height or 0)) / barOptions.count
	end

	for i = 1, self.barCount do
		if self.bars[i] then
			self.bars[i]:SetOptions(self.frame, self, i)
		else
			self.bars[i] = UIBar:new(self.frame, self, i)
		end
	end
	for i = self.barCount+1, #self.bars do
		self.bars[i]:Hide()
	end
end

function UIWindow.prototype:GetBarIndex(bar)
	for i = 1, self.barCount do
		if self.bars[i] == bar then
			return i
		end
	end
	error("Internal error: Bar not found")
end

function UIWindow.prototype:PopDisplaySetStack(windowOptions)
	AssessmentBarInterface:PopWindowDisplaySet(self)
end

local windowMenu 

local channelExcludeList =
{
	[L["General"]] = true,
	[L["Trade"]] = true,
	[L["LocalDefense"]] = true,
	[L["WorldDefense"]] = true,
	[L["GuildRecruitment"]] = true,
}
local function InsertWindowMenuChannels(windowIndex, ...)
	for i=2, select('#', ...), 2 do
		local channelName = select(i, ...)
		if not channelExcludeList[channelName] then
			local channel = AssessmentBarInterface:NewTable()
			local passValue = AssessmentBarInterface:NewTable()
			local passValue2 = AssessmentBarInterface:NewTable()
			local channelIndex = select(i-1, ...)
			local chatInfo = ChatTypeInfo["CHANNEL"..channelIndex]
			
			channel.type = "execute"
			channel.name = ("|cff%02x%02x%02x%d. %s|r"):format(chatInfo.r*255, chatInfo.g*255, chatInfo.b*255, channelIndex, channelName)
			channel.desc = channelName
			channel.func = windowReport
			
			passValue2[1] = "CHANNEL"
			passValue2[2] = channelName
			passValue[1] = windowIndex
			passValue[2] = passValue2
			channel.passValue = passValue
			
			windowMenu.args.report.args.channels.args[channelName] = channel
		end
	end
	
	if not next(windowMenu.args.report.args.channels.args) then
		windowMenu.args.report.args.channels = AssessmentBarInterface:ReleaseTable(windowMenu.args.report.args.channels, true)
	end
end

local function InsertWindowMenuDisplaySets(windowIndex)
	for k,v in AssessmentDisplaySets:IterateDisplaySets() do
		if v.selectable then
			table.insert(windowMenu.args.displaySet.validate, k)
		end
	end
	
	table.sort(windowMenu.args.displaySet.validate)
end

local function InsertWindowMenuFightData(windowIndex)
	for k,v in AssessmentDisplaySets:IterateFightData() do
		if type(k) == "number" and not windowMenu.args.fightData.args["history"..k] then
			local history = AssessmentDisplaySets:NewTable()
			local passValue = AssessmentDisplaySets:NewTable()
			
			history.type = "toggle"
			history.name = (L["History #%d"]):format(k)
			history.desc = history.name
			history.get = windowFightDataGet
			history.set = windowFightDataSet
			history.order = 3+k

			passValue[1] = windowIndex
			passValue[2] = k
			history.passValue = passValue

			windowMenu.args.fightData.args["history"..k] = history
		elseif type(k) == "string" then
		
			if not windowMenu.args.fightData.args.spacer2 then
				local spacer = AssessmentDisplaySets:NewTable()
				spacer.type = "header"
				spacer.order = 50
				windowMenu.args.fightData.args.spacer2 = spacer
			end
		
			-- Saved data
			local data = AssessmentDisplaySets:NewTable()
			local passValue = AssessmentDisplaySets:NewTable()
			
			data.type = "toggle"
			data.name = k
			data.desc = k
			data.get = windowFightDataGet
			data.set = windowFightDataSet

			passValue[1] = windowIndex
			passValue[2] = k
			data.passValue = passValue

			windowMenu.args.fightData.args[k] = data
		end
	end
end

local function DewdropChildrenMethod()
	Dewdrop:FeedAceOptionsTable(windowMenu)
end

function UIWindow.prototype:OnClick(buttonName)
	local windowIndex = AssessmentBarInterface:GetWindowIndex(self)
	if windowMenu then
		windowMenu = AssessmentBarInterface:ReleaseTable(windowMenu, true)
	end
	-- Use a window index as a parameter, rather than our own table, otherwise freeing it releases our tables.
	windowMenu = AssessmentBarInterface:CopyOptionsTemplate(windowMenuTemplate, windowIndex)
	InsertWindowMenuChannels(windowIndex, GetChannelList())
	InsertWindowMenuDisplaySets(windowIndex)
	InsertWindowMenuFightData(windowIndex)
	Dewdrop:Open("",
					'children', DewdropChildrenMethod,
					'cursorX', true,
					'cursorY', true
					)
end

function UIWindow.prototype:OnUpdateEvent()
	self.forceUpdate = true
end

function UIWindow.prototype:Show()
	if self.updater then
		self:CancelScheduledEvent(self.updater)
	end

	local windowOptions = self.windowOptions
	local updateRate = windowOptions.updateRate ~= 0 and 1/windowOptions.updateRate or 0

	self.updater = "Assessment-BarInterface-" .. tostring(self.frame)
	self:ScheduleRepeatingEvent(self.updater, self.Update, updateRate, self)
	self.frame:Show()
	self:Update()
	
	self:UnregisterAllEvents()
	for k in pairs(self.dependentEvents) do
		self:RegisterEvent(k, "OnUpdateEvent")
	end
end

function UIWindow.prototype:Hide()
	if self.updater then
		self:CancelScheduledEvent(self.updater)
		self.updater = nil
	end
	self.frame:Hide()
	self:UnregisterAllEvents()
end

function UIWindow.prototype:MergeRequiredData(data, events, overrideEvents)
	if data then
		if type(data) == "function" then
			data = data(unpack(self.listDataParameters))
		end
		
		if type(data) == "string" then
			self.dependentData[data] = true
		elseif type(data) == "table" then
			for k in pairs(data) do
				self.dependentData[k] = true
			end
		end
	end
	
	if events then
		if type(events) == "string" then
			self.dependentEvents[events] = true
		elseif type(events) == "table" then
			if #events > 0 then
				for i =1,#events do
					self.dependentEvents[events[i]] = true
				end
			else
				for k in pairs(events) do
					self.dependentEvents[k] = true
				end
			end
		end
	end
	
	if overrideEvents then
		if type(overrideEvents) == "string" then
			self.overrideEvents[overrideEvents] = true
		elseif type(overrideEvents) == "table" then
			if #overrideEvents > 0 then
				for i =1,#overrideEvents do
					self.overrideEvents[overrideEvents[i]] = true
				end
			else
				for k in pairs(overrideEvents) do
					self.overrideEvents[k] = true
				end
			end
		end
	end
end

function UIWindow.prototype:NeedsUpdate(fightData)
	if self.dependentData.ALWAYSUPDATE then
		return true
	end

	local dataUpdateTimes = self.dataUpdateTimes
	if not fightData then
		if next(dataUpdateTimes) then
			for k in pairs(dataUpdateTimes) do
				dataUpdateTimes[k] = nil
			end
			return true
		end
		if self.forceUpdate then
			self.forceUpdate = nil
			return true
		end
		return false
	end

	local result = false
	for k in pairs(self.dependentData) do
		if dataUpdateTimes[k] ~= fightData.dataUpdateTimes[k] then
			result = true
			dataUpdateTimes[k] = fightData.dataUpdateTimes[k]
		end
	end
	if self.forceUpdate then
		self.forceUpdate = nil
		return true
	end
	return result
end

local sortList = { }
local maximumBarValue
local totalBarValue

function UIWindow.prototype:UpdateSortList(fightData)
	maximumBarValue = 0
	totalBarValue = 0
	for i,v in ipairs(sortList) do
		sortList[i] = AssessmentBarInterface:ReleaseTable(v)
	end

	if fightData then
		local barValueMethod, filterMethod = self.barValueMethod, self.filterMethod
	
		local data = AssessmentBarInterface:NewTable()
		data.fightData = fightData
		data.window = self
		for k,v in self.listTypeMethod(fightData, unpack(self.listDataParameters)) do
			data.key = k
			data.value = v
			
			local barValue = barValueMethod(data)
			data.barValue = barValue
			
			if filterMethod(data) then
				if barValue > maximumBarValue then
					maximumBarValue = barValue
				end
				totalBarValue = totalBarValue + barValue

				local rank = #sortList+1
				data.rank = rank
				sortList[rank] = data
				
				data = AssessmentBarInterface:NewTable()
				data.fightData = fightData
				data.window = self
			end			
		end
		AssessmentBarInterface:ReleaseTable(data)
		
		for _,data in ipairs(sortList) do
			data.totalBarValue = totalBarValue
			data.maximumBarValue = maximumBarValue
			data.sortValue = self.sortMethod(data)
		end

		table.sort(sortList, self.sortDirectionMethod)
	end
end

Assessment.TagData.FightIndex =
{
	tableData = "window.fightIndex",
	guaranteeResult = true,
}

Assessment.TagData.WindowParameter =
{
	method = function(data, value, index)
				return value and data.window.windowParameters[index]
			 end,
	parameterTypes = { "number" }
}

Assessment.TagData.WindowParameterList =
{
	method = function(data, value, index)
				return value and table.concat(data.window.windowParameters, ",")
			 end,
}

Assessment.TagData.ListParameter =
{
	method = function(data, value, index)
				return value and data.window.listDataParameters[index]
			 end,
	parameterTypes = { "number" }
}

local titleData = { key = UnitName("player"), rank = 0 }
function UIWindow.prototype:DoUpdate(fightData)
	self:UpdateSortList(fightData)

	local barNormalizeMultiplier = maximumBarValue ~= 0 and 1 / maximumBarValue or 0

	local topBarOffset = self.topBarOffset
	local barCount = self.barCount
	if topBarOffset + barCount > #sortList then
		topBarOffset = #sortList - barCount
		if topBarOffset < 0 then
			topBarOffset = 0
		end
		self.topBarOffset = topBarOffset
	end

	local endDisplay = barCount
	if endDisplay > #sortList-topBarOffset then 
		endDisplay = #sortList-topBarOffset 
		for i = endDisplay+1, barCount do
			self.bars[i]:Hide()
		end
	end
	for i = 1, endDisplay do
		local rank = i+topBarOffset
		local data = sortList[rank]
		data.rank = rank
		self.bars[i]:Update(data, barNormalizeMultiplier, self)
	end	

	if self.windowOptions.title.enable then
		titleData.fightData = fightData
		titleData.totalBarValue = totalBarValue
		titleData.maximumBarValue = maximumBarValue
		titleData.window = self

		self.titleText:SetText( self.titleTextMethod(titleData) )
	end
end

function UIWindow.prototype:Update()
	local fightData = Assessment:GetFightData(self.fightIndex)
	if not self:NeedsUpdate(fightData) then return end
	self:DoUpdate(fightData)
end

function UIWindow.prototype:OnBarClicked(bar)
	if self.displaySet.onClickType == "SHOW_DISPLAY_SET" and self.displaySet.onClickDisplaySet and self.displaySet.onClickData then
		local barIndex = self:GetBarIndex(bar)

		self:UpdateSortList(Assessment:GetFightData(self.fightIndex))
		local data = sortList[barIndex+self.topBarOffset]
		
		if data then
			local displaySet = Assessment:CompileTag(self.displaySet.onClickDisplaySet, "string").method(data)
			local parameters = Assessment:CompileTag(self.displaySet.onClickData, "string").method(data)

			local displaySetAndParameters = displaySet.."$"..parameters

			if IsShiftKeyDown() then
				AssessmentBarInterface:NewWindowDisplaySet(self, displaySetAndParameters)
			else
				AssessmentBarInterface:PushWindowDisplaySet(self, displaySetAndParameters)
			end
		end
	end
end

local function StripColorCodes(s)
	if type(s) == "string" then
		return s:gsub("|c%x%x%x%x%x%x%x%x", ""):gsub("|r", "")
	else
		return s
	end
end


function UIWindow.prototype:Report(chatType, limit, channel)
	channel = chatType == "CHANNEL" and GetChannelName(channel) or channel

	local fightData = Assessment:GetFightData(self.fightIndex)
	self:UpdateSortList(fightData)
	
	if self.windowOptions.title.enable then
		titleData.fightData = fightData
		titleData.isReport = true
		titleData.totalBarValue = totalBarValue
		titleData.maximumBarValue = maximumBarValue
		titleData.window = self
		
		if chatType == "CHATFRAME" then
			_G[channel]:AddMessage(StripColorCodes(self.titleTextMethod(titleData)))
		else
			SendChatMessage(StripColorCodes(self.titleTextMethod(titleData)), chatType, nil, channel)
		end
		titleData.isReport = nil
	end
	
	if limit and limit ~= 0 then 
		if #sortList < limit then
			limit = #sortList
		end
	else
		limit = #sortList
	end

	for i = 1,limit do
		local text = AssessmentBarInterface:NewTable()
		local data = sortList[i]
		data.rank = i
		data.isReport = true
		
		for j = 1,self.displaySet.text.count do
			local t = StripColorCodes(self.textMethods[j](data))
			if t ~= "" then
				text[#text+1] = t
			end
		end
		
		if chatType == "CHATFRAME" then
			_G[channel]:AddMessage(table.concat(text, ", "))
		else
			SendChatMessage(table.concat(text, ", "), chatType, nil, channel)
		end
		AssessmentBarInterface:ReleaseTable(text)
	end
end

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------

function AssessmentBarInterface:CreateWindowOptions(i)
	local options = self:CopyOptionsTemplate(windowOptionsTemplate, i)
	options.name = L["Window %d"]:format(i)
	options.desc = L["Options for assessment window %d"]:format(i)
	return options
end

function AssessmentBarInterface:UpdateDisplaySetValidate()
	for k in pairs(displaySetValidate) do
		displaySetValidate[k] = nil
	end
	for k, v in AssessmentDisplaySets:IterateDisplaySets() do
		if v.selectable then
			displaySetValidate[k] = k
		end
	end
end

function AssessmentBarInterface:CreateOptions()
	self:ReleaseTable(Assessment.Options.args.windows.args, true)
	Assessment.Options.args.windows.args = self:CopyOptionsTemplate(windowsArgs)

	for i = 1, self.db.profile.windows.count do
		if Assessment.Options.args.windows.args[tostring(i)] then
			self:ReleaseTable(Assessment.Options.args.windows.args[tostring(i)], true)
		end
		Assessment.Options.args.windows.args[tostring(i)] = self:CreateWindowOptions(i)
	end
end

function AssessmentBarInterface:RemoveWindow(index)
	local windowOptions = self.db.profile.windows
	local windows = self.windows

	local oldWindow = windows[index]
	local oldWindowOptions = windowOptions[index]

	for i = index, windowOptions.count-1 do
		windowOptions[i] = windowOptions[i+1]
		windows[i] = windows[i+1]
	end
	windows[windowOptions.count] = oldWindow
	windowOptions[windowOptions.count] = oldWindowOptions
	windowOptions.count = windowOptions.count - 1
	
	oldWindow:Hide()	
	self:CreateOptions()
end

function AssessmentBarInterface:UpdateWindowOptions(index)
	self.windows[index]:SetWindowOptions(self.db.profile.windows[index])
end

local reportChatTypeLookup = { G = "GUILD", H = "CHANNEL", O = "OFFICER", P = "PARTY", R = "RAID", S = "SAY", W = "WHISPER", C = "CHATFRAME" }
local lastReportWindowIndex = 1
local function AssessmentReport(message)
	local _, _, windowIndex, chatType, limit, channel = string.find(message, "[ ]*(%d*)([a-zA-Z])(%d*)[ ]*(.*)");
	
	local reportChatType = reportChatTypeLookup[string.upper(chatType or "")]
	if not reportChatType or ((reportChatType == "WHISPER" or reportChatType == "CHANNEL") and channel == "") then
		for i=1,16 do
			DEFAULT_CHAT_FRAME:AddMessage(L["ReportUsage"..i])
		end
	else
		if reportChatType == "CHATFRAME" then
			channel = tonumber(channel) or 1
			if channel < 1 or channel > 7 then channel = 1 end
			channel = "ChatFrame"..channel
		end
	
		lastReportWindowIndex = tonumber(windowIndex) or lastReportWindowIndex
		limit = tonumber(limit)
		AssessmentBarInterface:Report(lastReportWindowIndex, reportChatType, limit, channel)
	end
end

function AssessmentBarInterface:OnInitialize()
	self.windows = { }

	if not self.db.profile.windows.count then
		self.db.profile.windows.count = 1
		self.db.profile.windows[1] = self:CopyTable(defaultWindowProfile)
	end
	
	SLASH_ASR1 = "/asr"
	SlashCmdList["ASR"] = AssessmentReport
end

function AssessmentBarInterface:OnEnable()
	self:UpdateDisplaySetValidate()
	self:RegisterEvent("SharedMedia_SetGlobal", "UpdateAllWindows")
	self:RegisterEvent("AssessmentDisplaySets:OnDisplaySetAdded", "OnDisplaySetListChanged")
	self:RegisterEvent("AssessmentDisplaySets:OnDisplaySetRemoved", "OnDisplaySetListChanged")
	self:RegisterEvent("AssessmentDisplaySets:OnDisplaySetListChanged", "OnDisplaySetListChanged")
	self:RegisterEvent("AssessmentDisplaySets:OnDisplaySetRequestRemove", "OnDisplaySetRequestRemove")
	self:RegisterEvent("AssessmentDisplaySets:OnDisplaySetUpdated", "OnDisplaySetUpdated")
	self:RegisterEvent("Assessment:RedrawAll", "RedrawAll")
	self:RegisterEvent("Assessment:ResetFightData", "OnResetFightData")

	self:RegisterEvent("Assessment:TestVisible", "TestVisible")
	self:RegisterEvent("Assessment:HideAll", "HideAll")
	self:RegisterEvent("Assessment:ShowAll", "ShowAll")

	self:CreateWindows()
	self:CreateOptions()
end

function AssessmentBarInterface:OnDisable()
	for i = 1, self.db.profile.windows.count do
		self.windows[i].forceUpdate = true
	end
end

function AssessmentBarInterface:UpdateAllWindows()
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		self.windows[i]:SetWindowOptions(windowOptions)
	end
end

function AssessmentBarInterface:CreateWindows()
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		if self.windows[i] then
			self.windows[i]:SetWindowOptions(windowOptions)
		else
			self.windows[i] = UIWindow:new(windowOptions)
		end
	end
	
	for i = self.db.profile.windows.count+1, #self.windows do
		self.windows[i]:Hide()
	end
end

function AssessmentBarInterface:GetWindowIndex(window)
	for i = 1,self.db.profile.windows.count do
		if self.windows[i] == window then
			return i
		end
	end
	error("INTERNAL ERROR: Window not found")
end

function AssessmentBarInterface:SetDisplaySet(window, displaySetName)
	Dewdrop:Close()
	local windowIndex = self:GetWindowIndex(window)
	local windowOptions = self.db.profile.windows[windowIndex]
	windowOptions.displaySetStackLevel = 1
	windowOptions.displaySetStack[1] = displaySetName
	self:UpdateWindowOptions(windowIndex)
end

function AssessmentBarInterface:PushWindowDisplaySet(window, displaySetName)
	local windowIndex = self:GetWindowIndex(window)
	local windowOptions = self.db.profile.windows[windowIndex]
	
	local displaySet = AssessmentDisplaySets:GetDisplaySet(("$"):split(displaySetName, 2))
	if not displaySet then
		error("Unable to set display set: "..displaySetName)
	end
	
	windowOptions.displaySetStackLevel = windowOptions.displaySetStackLevel + 1
	windowOptions.displaySetStack[windowOptions.displaySetStackLevel] = displaySetName
	self:UpdateWindowOptions(windowIndex)
end

function AssessmentBarInterface:PopWindowDisplaySet(window)
	local windowIndex = self:GetWindowIndex(window)
	local windowOptions = self.db.profile.windows[windowIndex]
	
	local displaySetStackLevel = windowOptions.displaySetStackLevel
	if displaySetStackLevel == 1 then return end
	
	windowOptions.displaySetStack[displaySetStackLevel] = nil
	windowOptions.displaySetStackLevel = displaySetStackLevel - 1
	self:UpdateWindowOptions(windowIndex)
end

function AssessmentBarInterface:GetDisplaySetName(windowOptions)
	return ("$"):split(windowOptions.displaySetStack[windowOptions.displaySetStackLevel], 2)
end

function AssessmentBarInterface:GetDisplaySet(windowOptions)
	local displaySetName, displaySetParameters = self:GetDisplaySetName(windowOptions)
	return AssessmentDisplaySets:GetDisplaySet(displaySetName), displaySetParameters
end

function AssessmentBarInterface:Report(windowIndex, chatType, limit, channel)
	if windowIndex >= 1 and windowIndex <= self.db.profile.windows.count then
		self.windows[windowIndex]:Report(chatType, limit, channel)
	end
end

function AssessmentBarInterface:AddWindow()
	local windowOptions = self.db.profile.windows
	windowOptions.count = windowOptions.count + 1
	
	if not windowOptions[windowOptions.count] then
		windowOptions[windowOptions.count] = self:CopyTable(defaultWindowProfile)
	end
	
	self:CreateWindows()
	self:CreateOptions()
end

function AssessmentBarInterface:NewWindowDisplaySet(window, displaySetName)
	local windowOptions = self.db.profile.windows
	windowOptions.count = windowOptions.count + 1
	
	if not windowOptions[windowOptions.count] then
		windowOptions[windowOptions.count] = self:CopyTable(defaultWindowProfile)
	end

	-- Since display sets make use of inherited fight data, copy the stack.	
	local windowIndex = self:GetWindowIndex(window)
	local sourceWindowOptions = windowOptions[windowIndex]
	local targetWindowOptions = windowOptions[windowOptions.count]
	
	targetWindowOptions.displaySetStackLevel = sourceWindowOptions.displaySetStackLevel
	for i = 1,sourceWindowOptions.displaySetStackLevel do
		targetWindowOptions.displaySetStack[i] = sourceWindowOptions.displaySetStack[i]
	end

	targetWindowOptions.displaySetStackLevel = targetWindowOptions.displaySetStackLevel + 1
	targetWindowOptions.displaySetStack[targetWindowOptions.displaySetStackLevel] = displaySetName
	targetWindowOptions.fightIndex = sourceWindowOptions.fightIndex
	
	self:CreateWindows()
	self:CreateOptions()
end

-----------------------------------------------------------------------------

function AssessmentBarInterface:OnDisplaySetUpdated(displaySetName)
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		if self:GetDisplaySetName(windowOptions) == displaySetName then
			self.windows[i]:SetWindowOptions(windowOptions)
		end
	end
end

function AssessmentBarInterface:OnDisplaySetRequestRemove(displaySetName)
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		
		for j = 1, windowOptions.displaySetStackLevel do
			if ("$"):split(windowOptions.displaySetStack[j], 2) == displaySetName then
				AssessmentDisplaySets:PreventDisplaySetRemoval(L["This display set cannot be removed currently because one or more windows require its data."])
				return
			end
		end
	end
end

function AssessmentBarInterface:RedrawAll()
	for i=1,self.db.profile.windows.count do
		self.windows[i].forceUpdate = true
	end
end

function AssessmentBarInterface:OnDisplaySetListChanged()
	self:UpdateDisplaySetValidate()
	self:CreateOptions()
end

function AssessmentBarInterface:OnResetFightData()
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		if not Assessment:GetFightData(windowOptions.fightIndex) then
			windowOptions.fightIndex = false
			self.windows[i]:SetWindowOptions(windowOptions)
		end
	end
end

-----------------------------------------------------------------------------

function AssessmentBarInterface:TestVisible(data)
	for i = 1, self.db.profile.windows.count do
		if self.db.profile.windows[i].frame.visible then
			data.result = true
			return
		end
	end
end

function AssessmentBarInterface:HideAll()
	for i = 1, self.db.profile.windows.count do
		self.db.profile.windows[i].frame.visible = false
		self.windows[i]:Hide()
	end
end

function AssessmentBarInterface:ShowAll()
	for i = 1, self.db.profile.windows.count do
		self.db.profile.windows[i].frame.visible = true
		self.windows[i]:Show()
	end
end

-----------------------------------------------------------------------------
